import itertools


class ChangedEntry:

    id_iter = itertools.count()

    def __init__(self, applied_function: str, column: str, row_index: int, value_before: str, value_after: str):
        self.id: int = next(self.id_iter)
        self.applied_function: str = applied_function
        self.column: str = column
        self.row_index: int = row_index
        self.value_before: str = value_before
        self.value_after: str = value_after


class ChangeLog:

    def __init__(self):
        self.entries: list[ChangedEntry] = []

    def add_entry(self, entry: ChangedEntry):
        self.entries.append(entry)

    def pretty_print(self):
        for entry in self.entries:
            print(f'[{entry.id}] {entry.applied_function}: '
                  f'{entry.column}[{entry.row_index}] | '
                  f'{entry.value_before} -> {entry.value_after}')



import os
from unittest import TestCase

import numpy as np
import pandas as pd

from cleandat.cleanup import unify_number_format, clean_unknown_entries, \
    remove_entries_with_inconsistent_datatypes


class Test(TestCase):

    TEST_DIR_PATH = os.path.dirname(os.path.realpath(__file__))

    df = pd.read_csv(os.path.join(TEST_DIR_PATH, "resources", 'test.csv'))

    def test_unify_number_format(self):
        df_clean = unify_number_format(self.df)
        self.assertEqual(1500000, np.average(df_clean['cell_count'][9:16]))
        self.assertEqual(120000000, df_clean['cell_count'][17])

    def test_clean_unknown_entries(self):
        df_clean = clean_unknown_entries(self.df)
        self.assertEqual(pd.isna(df_clean['birth_date'])[13], True)
        self.assertEqual(pd.isna(df_clean['birth_date'])[23], True)
        self.assertEqual(pd.isna(df_clean['birth_date'])[24], True)
        self.assertEqual(pd.isna(df_clean['freetext'])[22], True)
        self.assertEqual(pd.isna(df_clean['freetext'])[27], True)

    def test_remove_entries_with_inconsistent_datatypes(self):
        df_clean = remove_entries_with_inconsistent_datatypes(self.df, threshold=0.5)
        self.assertEqual(pd.isna(df_clean['sex'])[19], True)


import os
from unittest import TestCase

import pandas as pd

from cleandat.encoding import get_encoding, identify_descriptive_header_cells, \
    identify_descriptive_header_rows, get_column_encoding_schemes, encode_dataframe


class Test(TestCase):

    TEST_DIR_PATH = os.path.dirname(os.path.realpath(__file__))

    df = pd.read_csv(os.path.join(TEST_DIR_PATH, "resources", 'test.csv'))

    def test_get_encoding_return_encoding(self):
        case1 = '1=m'
        case2 = '1->m'
        case3 = '1 - m'
        case4 = '1 : m'
        case5 = '1= m'
        case6 = '2 -> female'
        case7 = '3 = not specified'
        self.assertTupleEqual(('m', 1), get_encoding(case1))
        self.assertTupleEqual(('m', 1), get_encoding(case2))
        self.assertTupleEqual(('m', 1), get_encoding(case3))
        self.assertTupleEqual(('m', 1), get_encoding(case4))
        self.assertTupleEqual(('m', 1), get_encoding(case5))
        self.assertTupleEqual(('female', 2), get_encoding(case6))
        self.assertTupleEqual(('not specified', 3), get_encoding(case7))

    def test_get_encoding_reverse_order(self):
        case1 = "m:1"
        self.assertTupleEqual(('m', 1), get_encoding(case1))

    def test_get_encoding_return_none(self):
        case1 = '7 R-CHOP'
        case2 = '1'
        self.assertEqual(None, get_encoding(case1))
        self.assertEqual(None, get_encoding(case2))

    def test_identify_descriptive_header_cells(self):
        headers_bool = identify_descriptive_header_cells(self.df)
        self.assertEqual(headers_bool['sex'][:1].all(), True)
        self.assertEqual(pd.isnull(headers_bool['sex'][2:8]).all(), True)
        self.assertEqual(headers_bool['sex'][9:16].all(), False)
        self.assertEqual(pd.isnull(headers_bool['sex'][17]), True)
        self.assertEqual(headers_bool['sex'][18:28].all(), False)
        self.assertEqual(headers_bool['categories'][0:3].all(), True)

    def test_identify_descriptive_header_rows(self):
        header_rows = identify_descriptive_header_rows(self.df)
        self.assertEqual(len(header_rows), 3)
        self.assertEqual(header_rows, [0, 1, 2])

    def test_get_column_encoding_schemes(self):
        schemes = get_column_encoding_schemes(self.df)
        # TODO: this includes a false positive at pos 3
        self.assertEqual(3, len(schemes))
        self.assertEqual(2, len(schemes['sex']))
        self.assertEqual(3, len(schemes['categories']))

    def test_encode_dataframe(self):
        encoding_scheme = {'sex': {'m': 1, 'f': 2}, 'categories': {'foo': 1, 'bar': 2, 'foobar': 3}}
        df_encoded = encode_dataframe(self.df, encoding_scheme)
        self.assertEqual(1, (df_encoded['sex'][27]))
        self.assertEqual(1, (df_encoded['sex'][28]))
        self.assertEqual(2, df_encoded['categories'][17])
        self.assertEqual(1, df_encoded['categories'][27])
        self.assertEqual(3, df_encoded['categories'][15])





from pandas import DataFrame

from cleandat.cleanup import drop_rows, drop_empty_columns, drop_empty_rows, clean_unknown_entries, \
    remove_entries_with_inconsistent_datatypes, unify_number_format
from cleandat.encoding import get_column_encoding_schemes, encode_dataframe, \
    identify_descriptive_header_rows
from cleandat.date import identify_date_columns, normalize_date_entries, decompose_date_entries


def find_encodings_and_encode_strings(df: DataFrame, drop_encoding_rows: bool = True) -> DataFrame:
    """Finds the encoding schemes of a dataframe, encodes the dataframe and removes the encoding rows.

    :param drop_encoding_rows: Whether rows containing the encoding description should be dropped afterwards,
    default true
    :param df: The dataframe to be encoded
    :return: The encoded dataframe without the rows which describe the encoding schemes
    """
    encodings = get_column_encoding_schemes(df)
    df = encode_dataframe(df, encodings)
    if drop_encoding_rows:
        encoding_rows = identify_descriptive_header_rows(df)
        df = drop_rows(df, encoding_rows)
    return df


def remove_empty_columns_and_rows(df: DataFrame) -> DataFrame:
    """Removes empty columns and rows from a dataframe.

    :param df: The dataframe to be cleaned
    :return: The cleaned dataframe
    """
    df = drop_empty_columns(df)
    df = drop_empty_rows(df)
    return df


def clean_date_entries(df: DataFrame, decompose_dates: bool = True, remove_unparsable=True) -> DataFrame:
    """Cleans and encodes date entries in a dataframe.

    :param decompose_dates: Whether date entries should be decomposed into _day, month, year columns, default true
    :param remove_unparsable: Whether unparsable date entries should be removed, default true
    :param df: The dataframe to be cleaned
    :return: The cleaned dataframe

    """
    date_columns = identify_date_columns(df)
    df = normalize_date_entries(df, date_columns, remove_unparsable=remove_unparsable)
    if decompose_dates:
        df = decompose_date_entries(df, date_columns)
    return df


def remove_inconsistencies(df: DataFrame, threshold = 0.1) -> DataFrame:
    """Removes inconsistencies from a dataframe and replaces them with NaN.

    Should be used after as a last step (especially after encoding steps) since it removes data.

    :param df: The dataframe to be cleaned
    :param threshold: The threshold for the ratio of non-allowed inconsistent entries in a column, default 0.1 - meaning
    when more than 10% of the entries in a column are inconsistent (e.g. containing string instead of int), they get
    removed
    :return: The cleaned dataframe
    """
    df = unify_number_format(df)
    df = clean_unknown_entries(df)
    df = remove_entries_with_inconsistent_datatypes(df, threshold=threshold)
    return df



import numpy as np
import pandas as pd
from pandas import DataFrame

from cleandat import constants


def get_encoding(string: str) -> (str, int):
    """ Checks whether a string could be meant as documentation for an encoding.

    Encoding may be divided by a combination of whitespaces and/or delimiters, e.g. for male/female encoding
    1=m, 1->m, 1 - m, 1 : m.

    :param string: The string that should be checked
    :return: The assumed key and value of the encoding, if the string is a code string, else None
    """
    possible_delimiters = constants.DELIMITERS
    if any(delimiter in string for delimiter in possible_delimiters):
        for delimiter in possible_delimiters:
            if delimiter in string:
                try:
                    key, value = string.split(delimiter)
                except ValueError:
                    return None
                # remove trailing and leading whitespaces
                key = key.strip()
                value = value.strip()
                if key.isnumeric() and isinstance(value, str):
                    return value, int(key)
                # in case the order is swapped, e.g. m=1 instead of 1=m
                if value.isnumeric() and isinstance(key, str):
                    return key, int(value)

    else:
        return None


def identify_descriptive_header_cells(df: DataFrame) -> DataFrame:
    """Identifies cells that are likely to contain an encoding scheme opposed to containing actual data.

    Heuristic: If an entry contains a number and a string divided by some kind of delimiter (that may be a whitespace,
    a colon, etc.) and the number is contained in the data of this column, it is likely a descriptive header.

    Example: '1= Age' or '1 : Age' or '1 Age' or '1=Age' or '1->Age' or '1_Age'

    :param df: The dataframe to be analyzed
    :return: boolean matrix with the same dimensions as the dataframe, where True indicates a descriptive header cell
    """
    is_header_matrix = pd.DataFrame(False, index=df.index, columns=df.columns)
    for column in df:
        for row_idx, entry in enumerate(df[column]):
            # first case: cell is emtpy
            if pd.isnull(entry):
                is_header_matrix.at[row_idx, column] = np.nan
                continue
            # second case: cell is not an encoding cell
            if get_encoding(str(entry)) is None:
                continue
            # third case: cell is an encoding cell
            key, value = get_encoding(str(entry))
            # check if the key is also contained in the data of this column
            if key is not None and value is not None and int(value) in df[column]:
                is_header_matrix.at[row_idx, column] = True
    return is_header_matrix


def get_column_encoding_schemes(df: DataFrame) -> dict:
    """ Gets the encoding schemes from a heuristic of descriptive header cells of a dataframe.

    :param df: The dataframe from which the encoding schemes should be extracted
    :return: For each column, a dictionary containing the encoding schemes
    """
    is_header_cell = identify_descriptive_header_cells(df)
    # ignore empty cells
    is_header_cell.replace(np.nan, False, inplace=True)
    schemas = {}
    for column in df:
        col_mappings = {}
        for entry in df[column][is_header_cell[column]]:
            if entry is not None and get_encoding(entry) is not None:
                key, value = get_encoding(entry)
                col_mappings[key] = value
        if len(col_mappings) > 0:
            schemas[column] = col_mappings
    return schemas


def identify_descriptive_header_rows(df: DataFrame, error_tolerance: float = 0.1) -> list[int]:
    """Identifies rows that are likely to contain an encoding scheme opposed to containing actual data.

    Heuristic: if a row contains a descriptive header cell and only other entries that are either non-empty or other
    descriptive header entries, it is likely that the whole row is used to describe an encoding of the data.

    :param df: The dataframe to be analyzed
    :param error_tolerance: Percentage of entries allowed that do not conform to the heuristic
    :return: Row indices of rows that are likely to contain an encoding scheme
    """
    bool_headers_df = identify_descriptive_header_cells(df)
    row_indices = []
    for idx, row in enumerate(bool_headers_df.index):
        # check for rows containing encoding cells
        if bool_headers_df.loc[row].any():
            if (bool_headers_df.iloc[idx] == False).sum() / df.shape[1] < error_tolerance:
                row_indices.append(idx)
    return row_indices


def encode_dataframe(df: DataFrame, encoding_schemes: dict) -> DataFrame:
    """Encodes a dataframe according to a given encoding scheme.

    :param df: The dataframe to be encoded
    :param encoding_schemes: The encoding schemes to be used for encoding the dataframe
    :return: The encoded dataframe
    """
    encoded_df = df.copy()
    for column in df:
        if column in encoding_schemes:
            encoded_df[column] = encoded_df[column].replace(encoding_schemes[column])
    return encoded_df



import os
from datetime import datetime
from unittest import TestCase

import pandas as pd

from cleandat.date import identify_date_columns, normalize_date_entries, decompose_date_entries, \
    create_durational_column


class Test(TestCase):

    TEST_DIR_PATH = os.path.dirname(os.path.realpath(__file__))

    df = pd.read_csv(os.path.join(TEST_DIR_PATH, "resources", 'test.csv'))

    def test_identify_date_columns(self):
        date_columns = identify_date_columns(self.df.copy())
        self.assertEqual(len(date_columns), 1)
        self.assertEqual(date_columns[0], 'birth_date')

    def test_clean_date_entries(self):
        df_cleaned = normalize_date_entries(self.df.copy(), ['birth_date'])
        self.assertEqual(pd.isna(df_cleaned['birth_date'])[10], True)
        self.assertEqual(df_cleaned['birth_date'][9], datetime(2020, 4, 12, 0, 0))
        self.assertEqual(df_cleaned['birth_date'][11], datetime(2021, 6, 4, 0, 0))
        self.assertEqual(df_cleaned['birth_date'][12], datetime(1990, 2, 2, 0, 0))
        self.assertEqual(df_cleaned['birth_date'][26], datetime(1999, 9, 9, 0, 0))

    def test_decompose_date_entries(self):
        df_dates_clean = pd.read_csv(os.path.join(os.path.dirname(os.path.realpath(__file__)), "resources",
                                                  'test_dates_cleaned.csv'))
        df_decomposed = decompose_date_entries(df_dates_clean, ['birth_date'])
        self.assertEqual(df_decomposed['birth_date_year'][9], 2020)
        self.assertEqual(df_decomposed['birth_date_month'][9], 4)
        self.assertEqual(df_decomposed['birth_date_day'][9], 12)
        self.assertEqual(df_decomposed['birth_date_year'][11], 2021)
        self.assertEqual(df_decomposed['birth_date_month'][11], 6)
        self.assertEqual(df_decomposed['birth_date_day'][11], 4)
        self.assertEqual(df_decomposed['birth_date_year'][12], 1990)
        self.assertEqual(df_decomposed['birth_date_month'][12], 2)
        self.assertEqual(df_decomposed['birth_date_day'][12], 2)
        self.assertEqual(df_decomposed['birth_date_year'][26], 1999)
        self.assertEqual(df_decomposed['birth_date_month'][26], 9)
        self.assertEqual(df_decomposed['birth_date_day'][26], 9)


import dateparser
import numpy as np
import pandas as pd
from pandas import DataFrame


def identify_date_columns(df: DataFrame, threshold: float = 0.5) -> list:
    """Identifies columns that are likely to contain date entries.

    Heuristic: if a column contains a large number of date entries, it is likely that the column is a date column
    (e.g. a column containing the date of birth of a person).

    :param df: The dataframe to be analyzed
    :param threshold: The percentage of entries that have to be dates for a column to be considered a date column
    :return: A list of column names that are likely to contain date entries
    """
    date_columns = []
    for column in df:
        if df[column].dtype == 'object':
            pro_date_heuristic = 0
            for entry in df[column]:
                if str(entry).isnumeric() or pd.isnull(entry):
                    continue
                else:
                    if dateparser.parse(str(entry)) is not None:
                        # entry can be interpreted as a date
                        pro_date_heuristic += 1
            # take empty cell (nan) values out of the consideration
            num_nan = df[column].isna().sum()
            if pro_date_heuristic > threshold * (len(df[column]) - num_nan):
                date_columns.append(column)
    return date_columns


def create_durational_column(df: DataFrame, date_col_start: str, date_col_end: str, new_col_name: str, remove_dates: bool = True) -> DataFrame:
    """Creates a new column containing the duration between two date columns in days.

    :param df: The dataframe to which the column should be added
    :param date_col_start: The column containing the start date
    :param date_col_end: The column containing the end date
    :param new_col_name: The name of the new column
    :param remove_dates: If True, remove the date columns after the new column has been created
    :return: The dataframe with the new column
    """
    df[new_col_name] = (df[date_col_end] - df[date_col_start]).dt.days
    if remove_dates:
        df.drop([date_col_start, date_col_end], axis=1, inplace=True)
    return df


def normalize_date_entries(df: DataFrame, date_columns: list[str], date_order: str = 'DMY',
                           remove_unparsable=True) -> DataFrame:
    """Replace all date entries with a uniform format.

    Reformat all date entries to a uniform format - dates may be written in different formats,
    e.g. 01.01.2020 or January 1st, 2020

    :param remove_unparsable: If True, replace entries that cannot be parsed as dates with NaN
    :param date_order: The order of the date entries, e.g. DMY for 01.01.2020 or MDY for January 1st, 2020
    :param df: The dataframe to be reformatted
    :param date_columns: List of column names containing date entries
    :return: The reformatted dataframe
    """
    for col in date_columns:
        df[col] = df[col].apply(lambda x: dateparser.parse(str(x), settings={'DATE_ORDER': date_order})
        if not pd.isnull(x) and dateparser.parse(str(x), settings={'DATE_ORDER': date_order}) is not None else np.nan
        if remove_unparsable else x)
    return df


def decompose_date_entries(df: DataFrame, date_columns: list[str]) -> DataFrame:
    """Decomposes date entries into year, month and day.

    :param df: The dataframe to be decomposed
    :param date_columns: List of column names containing date entries
    :return: The decomposed dataframe
    """
    for col in date_columns:
        df[col + '_year'] = pd.DatetimeIndex(df[col]).year
        df[col + '_month'] = pd.DatetimeIndex(df[col]).month
        df[col + '_day'] = pd.DatetimeIndex(df[col]).day
        df.drop(col, axis=1, inplace=True)
    return df



import logging
from typing import Callable

from pandas import DataFrame

from cleandat.changelog import ChangedEntry, ChangeLog


class TransformationPipeline:

    def __init__(self, df: DataFrame):
        self.steps: list[Callable[[DataFrame, list[str]], DataFrame]] = []
        self.changelog: ChangeLog = ChangeLog()
        self.data: DataFrame = df

    def add_task(self, task: Callable[[DataFrame, list[str]], DataFrame], columns: list[str] = None):
        self.steps.append((task, columns))

    def run(self):
        for task, columns in self.steps:
            before_transformation = self.data.copy()
            logging.info(f'Running task {task} on columns {columns}')
            self.data = task(self.data, columns)
            self._extend_changelog(task, before_transformation, self.data)
        return self.data

    def _extend_changelog(self, transformation_step: Callable, df_before: DataFrame, df_after: DataFrame):
        diff = df_before.compare(df_after)
        different_columns = {x[0] for x in diff.columns}
        for column in different_columns:
            for row_idx in diff[column].index:
                value_before = str(diff[column].loc[row_idx, 'self'])
                value_after = str(diff[column].loc[row_idx, 'other'])
                function_name = transformation_step.__name__
                log_entry = ChangedEntry(function_name, column, row_idx, value_before, value_after)
                self.changelog.add_entry(log_entry)

    def print_changelog(self):
        return self.changelog.pretty_print()



import os
from unittest import TestCase

import pandas as pd

from cleandat.pipeline import TransformationPipeline
from cleandat.date import normalize_date_entries


class Test(TestCase):

    TEST_DIR_PATH = os.path.dirname(os.path.realpath(__file__))

    df = pd.read_csv(os.path.join(TEST_DIR_PATH, "resources", 'test.csv'))

    def test_pipeline_unknown_function(self):
        pipeline = TransformationPipeline(self.df)
        pipeline.add_task(normalize_date_entries, ['birth_date'])
        pipeline.run()
        pipeline.print_changelog()


import re
import numpy as np
import pandas as pd
from pandas import DataFrame

from cleandat.constants import MISSING_DATA_TOKENS


def clean_unknown_entries(df: DataFrame) -> DataFrame:
    """Takes a dataframe as input and returns the same dataframe with all entries that are unknown replaced by NaN.

    :param df: DataFrame: Specify the dataframe that is passed into the function
    :return: A dataframe
    """
    #  if it is not known, it should be either empty or nan
    tokens = MISSING_DATA_TOKENS
    df.replace(tokens, np.nan, inplace=True)
    return df


def drop_rows(df: DataFrame, row_indices: list[int]) -> DataFrame:
    """
    The drop_rows function takes a DataFrame and a list of row indices as input.
    It drops the rows in the DataFrame that correspond to those indices, and returns
    the resulting DataFrame.

    :param df: DataFrame: Specify the dataframe that will be passed into the function
    :param row_indices: list[int]: Specify the indices of rows to be dropped
    :return: A dataframe
    """
    df.drop(row_indices, axis=0, inplace=True)
    return df


def drop_empty_rows(df: DataFrame) -> DataFrame:
    """
    The drop_empty_rows function takes a dataframe as input and drops all rows that are empty.
        It returns the modified dataframe.

    :param df: Pass in the dataframe that we want to drop empty rows from
    :return: The dataframe with rows that are empty
    """
    df.dropna(axis=0, how='all', inplace=True)
    return df


def drop_empty_columns(df: DataFrame) -> DataFrame:
    """
    The drop_empty_columns function takes a dataframe as an argument and returns the same dataframe with any columns
    that are entirely empty (i.e., contain only NaN values) removed.

    :param df: Pass in the dataframe that we want to drop columns from
    :return: A dataframe with all columns that are empty dropped
    """
    for columns in df:
        if df[columns].isnull().all():
            df.drop(columns, axis=1, inplace=True)
    return df


def remove_entries_with_inconsistent_datatypes(df: DataFrame, threshold: float = 0.1) -> DataFrame:
    """ Removes entries that seem inconsistent with the rest of the column (e.g. string values in columns containing
    >90% numbers) and replace them with NaN.

    Only execute this after fully encoding the data / finishing all previous cleaning steps.
    :param df: The dataframe to be cleaned
    :param threshold: The threshold for the ratio of non-allowed inconsistent entries in a column, default 0.1 - meaning
    when more than 10% of the entries in a column are inconsistent (e.g. containing string instead of int), they get
    removed
    :return: The cleaned dataframe
    """
    for column in df:
        if df[column].dtype == 'object':

            # exclude nan entries from the calculation
            number_nan = sum([pd.isna(entry) for entry in df[column]])

            # calculate the percentage of numeric entries in the column
            percentage_numeric = sum([str(entry).isnumeric() for entry in df[column]]) / (len(df[column]) - number_nan)

            # numeric entries which are smaller than the threshold are considered inconsistent
            if percentage_numeric < threshold:
                df[column] = df[column].apply(lambda x: np.nan if str(x).isnumeric() else x)

            # string entries which are smaller than the threshold are considered inconsistent
            if 1 - percentage_numeric < threshold:
                df[column] = df[column].apply(lambda x: np.nan if not str(x).isnumeric() else x)

    return df


def unify_number_format(df: DataFrame) -> DataFrame:
    """Unify the number format of all entries in the dataframe.

    Data may be presented in different formats, e.g. in scientific notation (3.453 * 10^-4).
    This function unifies the format to a simple decimal number.

    :param df: The dataframe to be cleaned
    :return: The cleaned dataframe
    """
    for column in df:
        # replace all entries that contain a comma with a dot
        df[column] = df[column].replace(',', '.', regex=True)

        # replace scientific notation string with machine-readable notation
        # first: replace unicode superscript numbers with regular integers (e.g. ⁴ -> ^4)
        df[column] = df[column].apply(lambda x: replace_unicode_superscript_numbers(x) if pd.notna(x) else x)
        # second: replace the scientific notation with machine-readable notation
        df[column] = df[column].replace('10\s*\^', 'e+0', regex=True)

        # multiplication signs (either x or *) can be replaced with a whitespace
        df[column] = df[column].replace('\s?[\*x]\s?', ' ', regex=True)

        # dashes most likely imply ranges, e.g. when something can't be measured precisely -> take average instead
        df[column] = df[column].apply(lambda x: replace_range_with_average(x) if pd.notna(x) else x)

        # replace exponential notation with float
        df[column] = df[column].apply(lambda x: convert_exponential_to_float(x.replace(" ", "")) if pd.notna(x) else x)

    return df


def replace_range_with_average(input_string: str) -> str:
    """ Takes a string as input and returns the average of the lower and upper bounds of the range

    :param input_string: The input string
    :return: The average of the lower and upper bounds of the range if the input string matches the pattern, #
    else the input string
    """
    if pd.notna(input_string):
        # Define a regular expression pattern to match the range format (e.g., 1-2 or 0.5-1.7)
        pattern = r"(\d+(\.\d+)?)-(\d+(\.\d+)?)"

        # Find all occurrences of the pattern in the input_string
        matches = re.findall(pattern, str(input_string))

        if matches:
            # Calculate the average of the lower and upper bounds of the range
            lower_bound = float(matches[0][0])
            upper_bound = float(matches[0][2])
            average = (lower_bound + upper_bound) / 2

            # Replace the matched range with the calculated average
            return re.sub(pattern, str(average), input_string)

    return input_string


def replace_unicode_superscript_numbers(input_string: str) -> str:
    """
    The replace_unicode_superscript_numbers function takes a string as input and returns the same string with any
    unicode superscript numbers replaced by their equivalent LaTeX code. For example, it will replace ² with ^2.

    :param input_string: Store the string that is passed into the function
    :return: A string with unicode superscript numbers replaced by
    """
    for sup, num in zip("⁰¹²³⁴⁵⁶⁷⁸⁹", "0123456789"):
        input_string = str(input_string).replace(sup, "^" + num)
    return input_string


def convert_exponential_to_float(value: str) -> float:
    """Converts a string containing an exponential notation to a float.

    :param value: Store the value of each cell in the dataframe
    :return: A float, or the original value if it can't be converted to a float
    :doc-author: Trelent
    """
    # Check if the value is a string containing an exponential notation
    if isinstance(value, str) and 'e' in value.lower():
        try:
            # Convert exponential notation to float
            return float(value)
        except ValueError:
            # If the conversion fails, keep the original value
            return value
    # Check if the value is NaN
    elif pd.isna(value):
        return np.nan
    else:
        return value



import os
import io

from setuptools import setup

DESCRIPTION = 'Python functions to facilitate the pre-processing of data for ML tasks in a clinical context.'

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

setup(
    name='cleandat',
    version='0.0.2',
    packages=['tests', 'cleandat'],
    url='https://github.com/tiadams/cleandat',
    license='MIT License',
    author='Tim Adams',
    author_email='tim-adams@gmx.net',
    description=DESCRIPTION,
    long_description=long_description,
)



import os
from unittest import TestCase

import pandas as pd

from cleandat.workflows import clean_date_entries


class Test(TestCase):

    TEST_DIR_PATH = os.path.dirname(os.path.realpath(__file__))

    df = pd.read_csv(os.path.join(TEST_DIR_PATH, "resources", 'test.csv'))

    def test_clean_date_entries(self):
        df_clean = clean_date_entries(self.df)
        self.assertEqual(pd.isna(df_clean['birth_date_year'])[10], True)
        self.assertEqual(pd.isna(df_clean['birth_date_year'])[13], True)
